/* library for handling (sorting) positions on sequences */
/* 2010 Takeaki Uno */

#ifndef _pos_c_
#define _pos_c_

#include"pos.h"
#include"itemset.c"
#include"seq.c"

QSORT_TYPE(POS_SC, POS_SC)

/* initialization of POS */
void POS_init (POS *P){
  P->flag = 0;
  
    // for sc
  P->sc = NULL;
  P->str_sc = NULL;
  P->sc_skip = POS_DEFAULT_SC_SKIP;
  P->sc_max = P->sc_th = POS_SC_END+1;
  P->str_sc_max = INTHUGE;
  P->str_sc_end = 0;
  P->sc_cnt1 = P->sc_cnt2 = P->sc_cnt_all = 0;
  P->init_str_sc = 0;
  P->init_sc = P->sc_mark = P->sc_mark2 = 0;

    // for radix sort
  P->let_cnt = 0;
  P->jump = NULL;
  P->dep_max = P->hash_end = 0;
  P->hash_buf_siz = POS_DEFAULT_HASH_BUFSIZ;
  P->pos_dep = NULL;
  P->IDcnt = P->IDcnt2 = NULL;
  P->IDjump.v = NULL;

  P->remain_block.v = NULL;

    // for position list
  P->fname = NULL;
  P->pos_list = NULL;
  P->pos_list_num = 0;
  P->skip_chr = SEQ_CHAR_END-1;
  P->skip_chr_ub = INTHUGE;
  P->skip_repeat_th = INTHUGE;
  P->skip_repeat_check = INTHUGE;

    // for partition
  P->ppartition.v = NULL;
  P->ppartition_mark = NULL;
  P->pblock_num = 1; P->pblock_num_org = 0;
  P->chunk = 1; P->chunk_org = 0;
  P->cur_chunk = P->chunk_head = P->chunk_tail = 0;
  P->chunk_pos = NULL;

  P->k_level = 0;
  P->sep_pos = 0;
  P->partition_th = 0;
}

/* initialization of SEQ. This should be called before loading the string data */
/* execution is SEQ_init => SEQ_scan_file => determing dep_max & pos_dep & hash_end
   => SEQ_init2 => SEQ_read_file */

/* before this, dep_max has to be determined independently (and, pos_dep) */
/* before this, partition has to be determined */

void POS_alloc (POS *P, SEQ *S, ITEMSET *I){
    // set pointer to the related structures
  P->S = S; P->I = I;

    // for position 
  calloc2 (P->sc_cnt1, S->err_max+2, goto ERR);
  calloc2 (P->sc_cnt2, S->err_max+2, goto ERR);
  calloc2 (P->sc_cnt_all, S->err_max+2, goto ERR);

    // allocate sc
  if ( P->flag & POS_SC_USE){
    calloc2 (P->sc, S->str_len/POS_DEFAULT_SC_SKIP+2, goto ERR);
    ARY_FILL (P->sc, 0, S->str_len/POS_DEFAULT_SC_SKIP+2, P->init_sc);

  }
    // allocate str_sc
  if ( P->flag & POS_STR_SC ){
    malloc2 (P->str_sc, S->str_num_ +2, goto ERR);
    ARY_FILL (P->str_sc, 0, S->str_num_+2, P->init_str_sc);
  }
  if ( P->k_level>0 ){ P->sc_max /= P->k_level; P->str_sc_max /= P->k_level; }

    // determine optimal paritition of pattern string to blocks
  S->perr_org2 = S->perr_org;
  P->chunk = P->chunk_org>0? P->chunk_org: POS_opt_superpartition (P);
  S->plen = ((S->plen_org+P->chunk-1)/P->chunk);
  S->perr = S->perr_org/P->chunk;
  P->pblock_num = P->pblock_num_org>0? P->pblock_num_org: POS_opt_partition (S, 1);
  POS_mk_ppartition (P, P->pblock_num);

  if ( P->flag & POS_REMAIN_BLOCK ){
    QUEUE_alloc (&P->remain_block, P->pblock_num*2+2);
    SEQ_set_pperm (S);
  }

  P->sep_pos = S->str_len;
  if ( S->output_pos>0 ) ENMIN (P->sep_pos, S->output_pos);
  if ( S->separated_pos > 0 ) ENMIN (P->sep_pos, S->separated_pos);

  return;
  ERR:;
  mfree (P->sc_cnt1, P->sc_cnt2, P->sc_cnt_all, P->sc, P->str_sc);
}

/* terminal routine for POS */
void POS_end (POS *P){
  int i;

  mfree (P->sc_cnt1, P->sc_cnt2, P->sc_cnt_all);
  if ( P->jump ) FLOOP (i, 0, P->dep_max+2) QUEUE_end (&P->jump[i]);
  mfree (P->sc, P->str_sc, P->pos_list, P->jump);
  
  free2d (P->let_cnt);
  QUEUE_end (&P->IDjump);
#ifdef POS_PPERM_END
  free2d (P->pperm);
#endif
  QUEUE_end (&P->ppartition);
  mfree (P->ppartition_mark, P->pos_dep, P->chunk_pos, P->IDcnt, P->IDcnt2);
  QUEUE_end (&P->remain_block);
}

/* allocate memory for hash table (letter count table) */
void POS_hash_alloc (POS *P){
  int i;
  malloc2 (P->jump, P->dep_max+2, goto ERR);
  calloc2d (P->let_cnt, P->dep_max+2, P->hash_end+2, goto ERR);
  FLOOP (i, 0, P->dep_max+2) QUEUE_alloc (&P->jump[i], P->hash_end+2);

   // computing pos_dep, remain_block
  malloc2 (P->pos_dep, P->dep_max*4+2, goto ERR);
  P->dep_max = 0; POS_mk_pos_dep (P, 0);
  if ( !(P->flag & POS_REMAIN_BLOCK) )
      QUEUE_alloc (&P->remain_block, P->pblock_num*2+2);
  SEQ_set_pperm (P->S);

  return;
  ERR:;
  POS_end (P);
}

    // compute optimal hash size
void POS_hash_siz (POS *P){
  SEQ *S = P->S;
  int t, i, ii=S->plen+1, j=1, jj=1;
  size_t p, pp=0;
  QUEUE_INT *x;

  for (p=S->chr_num ; p<P->hash_buf_siz/j ; j++,p*=S->chr_num){
//  for (p=S->chr_num ; p<=MAX(256, S->chr_num) ; j++,p*=S->chr_num){
    i = t = 0;
    MQUE_FLOOP (P->ppartition, x){
      x++;
      while ( t < *x ){
        i++;
        t = MIN(t+j, *x);
      }
    }
    if ( ((double)i)*p > (double)(P->hash_buf_siz) ) break;
    if ( i < ii ){ ii=i; pp=p; jj=j; }
  }
//  if ( jj==0 ) error ("out of hush-buffer memory\n", goto ERR);
  P->dep_max = ii; P->dep_skip = jj;
  P->hash_end = pp;
//  if ( P->hash_end > P->hash_buf_siz || P->hash_end <= 0 ) P->hash_end = P->hash_buf_siz;
  POS_hash_alloc (P);
}

/* copy the data and allocate memory for all arrays, for multi_processing */
void POS_multicore_dup (POS *P, POS *PP, int level){
  *P = *PP;

  P->jump = NULL; P->pos_dep = NULL; P->ppartition_mark = NULL;
  P->ppartition.v = P->remain_block.v = NULL;

    // copy letter-counter of ID no less than "level"
  malloc2 (P->let_cnt, P->dep_max+2, goto ERR);
//  memcpy (P->let_cnt, PP->let_cnt, sizeof(*(P->let_cnt))*(P->dep_max+2));
  calloc2 (P->let_cnt[level], (P->hash_end+2)*(P->dep_max+2-level), goto ERR);
  FLOOP (common_int, level, P->dep_max+1) P->let_cnt[common_int+1] = P->let_cnt[common_int] + P->hash_end+2;

    // copy jump queue of ID no less than "level"
  malloc2 (P->jump, P->dep_max+1, goto ERR);
//  memcpy (P->jump, PP->jump, sizeof(*(P->jump))*(P->dep_max+1));
  FLOOP (common_int, level, P->dep_max+1) QUEUE_alloc (&P->jump[common_int], P->hash_end+2);

  malloc2 (P->pos_dep, P->dep_max*4+2, goto ERR);
  memcpy (P->pos_dep, PP->pos_dep, sizeof(int)*(PP->dep_max*4+2));
  calloc2 (P->ppartition_mark, P->pblock_num*4+4, goto ERR);

  if ( P->flag & POS_REMAIN_BLOCK ) P->remain_block = QUEUE_dup_ (&PP->remain_block);
  P->ppartition = QUEUE_dup_ (&PP->ppartition);
  return;
  ERR:
  POS_multicore_end (P, level);
}

void POS_multicore_end (POS *P, int level){
  FLOOP (common_int, level, P->dep_max+1) QUEUE_end (&P->jump[common_int]);
  if ( P->flag & POS_REMAIN_BLOCK ) QUEUE_end (&P->remain_block);
  QUEUE_end (&P->ppartition);
  free2 (P->let_cnt[level]);
  mfree (P->let_cnt, P->pos_dep, P->ppartition_mark, P->jump);
}

// determine pos_dep 
void POS_mk_pos_dep (POS *P, int s){
  QUEUE_INT *x;
  int j=0, i=P->dep_max;
  for ( x=P->ppartition.v+s ; x < P->ppartition.v+P->ppartition.t ; x++){
    j = *x;
    x++;
    while ( j < *x ){
      P->pos_dep[i*2] = j;
      j = MIN(j+P->dep_skip, *x);
      P->pos_dep[i*2+1] = j;
      i++;
    }
  }
  P->dep_max = i;
}


/************************************************************************************/
/************************************************************************************/

/* compute the optimal partition according to maching prob. p */
/* set S->pblock to the optimal partition number */
/* if flag, use heuristic adjastment */ 
int POS_opt_partition (SEQ *S, int flag){
  int i=S->perr+2, j, ii, s;
  double pp, ll=0, C, m;
  
  for ( s=0,C=1 ; C<1000000 ; s++,C*=S->chr_num );
  if ( s>1 ) s--;
  if ( flag && S->str_len < 500 ) i=S->perr+1;  // magic number here
  if ( S->perr == 0 ) return (1);
//  else if ( flag && S->str_len < 50000 ) i=S->perr+2;   // magic number here
  else if ( S->perr > 0 ){
    FLOOP (ii, S->perr+1, MIN(S->plen,S->perr+5)){
      C=1; FLOOP (j, 1, S->perr+1) C = C * (ii-j+1) / j;
      pp = S->str_len_org;
//      pp = pp / sqrt(S->interleave_len_org);

//      FLOOP (j, 0, (S->plen*(ii-S->perr))*4/ii/5) pp *= S->match_prob;
      FLOOP (j, 0, (S->plen*(ii-S->perr))/ii) pp *= S->match_prob;
//      m = C*MAX(1,S->plen/s/ii) + pp*S->plen*MAX(S->perr, S->perr/s/ii);
      m = C + pp;
      if ( m < ll || ii == S->perr+1 ){ ll = m; i = ii; }
if ( pp<1 ){ ll=m; i=ii; break; }
    }
//    if ( flag && S->plen>i && S->perr<3 && S->chr_num>100 && S->str_len > 10000000 ) i++;
  }
  return (i);
}

/* make ppartition according to pblock_num */
void POS_mk_ppartition (POS *P, int num){
  int j;
  QUEUE_alloc (&P->ppartition, num*4+4);
  calloc2 (P->ppartition_mark, num*4+4, EXIT);
  FLOOP (j, 0, num){
    QUE_INS (P->ppartition, P->S->plen*j/num);
    QUE_INS (P->ppartition, P->S->plen*(j+1)/num);
  }
  malloc2 (P->chunk_pos, P->chunk+2, EXIT);
  FLOOP (j, 0, P->chunk+1) P->chunk_pos[j] = P->S->plen_org*j/P->chunk;
}

/* estimate the computational cost for multi-classification */
double POS_estimate_cost (POS *P, int b){
  SEQ *S = P->S;
  double cost1=S->str_len, cost2 = ((double)S->str_len)*S->str_len*100;  // magic number here 
  int i;
  FLOOP (i, 1, b-S->perr+1) cost1 = cost1*(b-i+1)/i;
  if ( S->interleave_len_org>1 ) cost2 = cost2/S->interleave_len_org;
  FLOOP (i, 0, S->plen*(b-S->perr)/b) cost2 *= S->match_prob;
  return (cost1 + cost2);
}

/* find optimal partition of long substring, to pattern strings */
int POS_opt_superpartition (POS *P){
  int i, ii=1, b;
  double cost, mcost;
  SEQ *S = P->S;
  
  mcost = POS_estimate_cost (P, POS_opt_partition(S, 0));
  FLOOP (i, 2, S->perr_org+1){
    S->plen = S->plen_org/i;
    S->perr = S->perr_org/i;
    cost = POS_estimate_cost (P, (b=POS_opt_partition(S, 0)) )*i;
    if ( cost<mcost ){ mcost = cost; ii = i; }
  }
  S->plen = S->plen_org;
  S->perr = S->perr_org;
  if ( ii<4 ) return (1);
  return (ii);
}

/************************************************************************************/
/************************************************************************************/


/* setting patern permutation */
void SEQ_set_pperm (SEQ *S){
#ifdef SEQ_PPERM_END
  int i, j, k=0;
  malloc2 (S->pperm, SEQ_PPERM_END, EXIT);
  malloc2d (S->pperm[i], SEQ_PPERM_END, S->plen, EXIT);
  FLOOP (j, 0, S->pblock_num)
    for (i=j ; i<S->plen ; i+=S->pblock_num){ S->pperm[0][k] = i; k++; }
  FLOOP (i, 0, S->plen) S->pperm[1][i] = S->plen-1-S->pperm[0][i];
#endif
}

/*********************************************************************************/
/*********************************************************************************/


/*  output sc's for all positions 
  if POS_SC_OUTPUT_ONLY_MAX, then output the position with sc[] = POS_SC_END-1 */
void POS_fprint_pos_sc (FILE *fp, POS *P){
  SEQ *S = P->S;
  int flag;
  SEQ_ID i;
  SEQ_LONG t, x, start=S->str_len, prv=S->str_len;
  if ( P->flag & POS_STR_SC ){
    FLOOP (i, (S->flag&SEQ_SEGMENT)? 0: 1, S->str_num_){
      flag = (P->flag&POS_SC_OUTPUT_ONLY_MAX)? P->str_sc[i] >= P->sc_th: P->str_sc[i] <= P->sc_th;  // check this position is to be output or not
      if ( flag ) fprintf (fp, SEQ_LONGF "%c" SEQ_IDF "\n", i+S->base1, S->separator, P->str_sc[i]);
    }
    return;
  } else if ( S->interleave_len_org > 1 && !(P->flag&POS_SC_COUNT_INTERVAL) ){
    qsort_SEQ_LONG (P->pos_list, P->pos_list_num, 1); 
    FLOOP (t, 0, P->pos_list_num){
      x = P->pos_list[t];
      flag = (P->flag & POS_SC_OUTPUT_ONLY_MAX)? P->sc[x/POS_DEFAULT_SC_SKIP] >= P->sc_th:
          P->sc[x/POS_DEFAULT_SC_SKIP] <= P->sc_th;  // check position to be output or not
      if ( flag ){
        fprintf (fp, SEQ_LONGF "%c%d   ", x+S->base1, S->separator, P->sc[x/POS_DEFAULT_SC_SKIP]);
        SEQ_fprint_interval (fp, S, x, x, 0);
        fprintf (fp, "\n");
      }
    }
    return;
  }
  FLOOP (i, 1, (S->output_by>0? S->output_by: S->str_num)){
    if ( S->tail[i] - S->head[i] < S->plen_org ) continue;
    FLOOP (x, S->head[i], S->tail[i]-S->plen_org){
      flag = (P->flag & POS_SC_OUTPUT_ONLY_MAX)? P->sc[x/POS_DEFAULT_SC_SKIP] >= P->sc_th:
          P->sc[x/POS_DEFAULT_SC_SKIP] <= P->sc_th;  // check position to be output or not
      if ( flag ){
        if ( S->flag & SEQ_OUTPUT_INTERVAL ){
          if ( x != prv+1 || x == S->tail[i]-S->plen_org-1 ){
            if ( start < S->str_len && prv-start+1 >= (SEQ_LONG)S->plen_lb && prv-start+1 <= (SEQ_LONG)S->plen_ub ){
              fprintf (fp, SEQ_LONGF "%c" SEQ_LONGF, start+S->base1, S->separator, prv+S->base1);
              SEQ_fprint_interval (fp, S, start, prv, 0);
              fprintf (fp, "\n");
            }
            start = x;
          }
          prv = x;
        } else {
          fprintf (fp, SEQ_LONGF "%c%d   ", x+S->base1, S->separator, P->sc[x/POS_DEFAULT_SC_SKIP]);
          SEQ_fprint_interval (fp, S, x, x, 0);
          fprintf (fp, "\n");

        }
      }
    }
  }
}

/* compute statistics of sc, str_sc */
void POS_compute_sc_static (POS *P, int err){
  SEQ_ID j;
  SEQ_LONG p, i;
  SEQ *S = P->S;
  P->sc_cnt1[err] = P->sc_cnt2[err] = P->sc_cnt_all[err] = 0;
  
  if ( P->flag & POS_STR_SC ){
    FLOOP (j, (S->flag&SEQ_SEGMENT)? 0: 1, S->str_num_){
      if ( P->str_sc[j] == P->init_str_sc ) continue;
      if ( P->flag & POS_STR_SC_COUNT ){
        if ( P->str_sc[j] > 0 ){
          P->sc_cnt_all[err] += P->str_sc[j];
          P->sc_cnt1[err]++;
          if ( P->flag & POS_SC_OUTPUT_ONLY_MAX ) P->str_sc[j] = P->str_sc_max;
        }
      } else if ( P->flag & POS_STR_SC_MIN ){
        if ( P->str_sc[j] == err ){
          P->sc_cnt_all[err]++;
          P->sc_cnt1[err]++;
        }
      }
    }
  } else {
    FLOOP (i, 0, P->pos_list_num){
      p = P->pos_list[i];
      if ( P->sc[p] == P->init_sc ) continue;
      if ( P->flag & POS_SC_COUNT ){
        if ( P->sc[p] > 0 ){
          P->sc_cnt_all[err] += P->sc[p];
          P->sc_cnt1[err]++;
          if ( P->flag & POS_SC_OUTPUT_ONLY_MAX ) P->sc[p] = P->sc_max;
        }
      } else if ( P->flag & POS_SC_MIN ){
        if ( P->sc[p] == err ){
          P->sc_cnt_all[err]++;
          P->sc_cnt1[err]++;
        }
      }
    }
  }
}


/* print statistics of sc, str_sc (str_sc if flag!=0)*/
void POS_fprint_sc_static (POS *P, int flag){
  SEQ *S = P->S;
  SEQ_LONG i, rem=0, all = (S->flag&SEQ_SEGMENT)? S->segment_end: (flag? S->str_num: P->pos_list_end);
  int j;
  print_mes (S, "#mismatches #matching pos(str) (%%) #ave. degree\n");
  if ( P->flag & POS_SC_USE ){
    FLOOP (j, S->perr_min, S->perr_org2+1){
      print_mes (S, "         %2d  %10d(%2.2f%%) %8.5f\n", j, (int)P->sc_cnt1[j], ((double)P->sc_cnt1[j])/all*100, ((double)P->sc_cnt_all[j])/P->sc_cnt1[j]);
    }
  }
 
  if ( P->flag & POS_STR_SC ){
      FLOOP (i, 0, all) if ( P->str_sc[i] == P->init_str_sc ) rem++;
  } else FLOOP (i, 0, P->pos_list_num) if ( P->sc[P->pos_list[i]] == P->init_sc ) rem++;
  print_mes (S, ">=%2d          "SEQ_LONGF"(%2.2f%%)\n", S->perr_org2, rem, ((double)rem)/all*100);
}

/*  output sc's for all strings 
  if POS_SC_OUTPUT_ONLY_MAX, then output the position with sc[] = POS_SC_END-1 */
void POS_fprint_str_sc (FILE *fp, POS *P){
  SEQ_ID i;
  FLOOP (i, 1, P->S->output_by>0? P->S->output_by: P->S->str_num)
      fprintf (fp, "%d\n", P->str_sc[i]);
}

/* print a position (called by POS_fprint_pair ) */
/* for original string positions */
void SEQ_fprint_position (FILE2 *fp, SEQ *S, SEQ_LONG a, char cc){
  SEQ_ID sa = SEQ_get_substID(S, a);
  SEQ_LONG ay=0;
  int flag_a = ((S->flag&SEQ_SEGMENT) && a<S->matrix_pos);

// normalize positions
  if ( flag_a ){
    a -= S->head[sa]; ay = a/S->clms[sa]; a = a%S->clms[sa];
      // if not multi-string comprison, segment ID is start from the top of input data for every string
//    if ( sa>1 && S->separated_by==0 && (S->normalized_by==0 || sa<(SEQ_ID)S->normalized_by))
//        ay += (S->tail[1]-S->head[1]+S->clms[1]-1)/S->clms[1];
  } else if ( S->normalized_by>0 && sa>=(SEQ_ID)S->normalized_by ) a -= S->head[S->normalized_by];
  else if ( S->flag & (SEQ_OUTPUT_ID+SEQ_NORMALIZE) ) a -= S->head[sa];
  else a += a>S->separated_pos? S->base2: S->base1;
  if ( S->flag&SEQ_OUTPUT_ID ){ FILE2_print_int (fp, sa+S->base1, cc); cc=S->separator; }
  if ( !(S->flag&SEQ_NO_OUTPUT_POSITION) ){ FILE2_print_int (fp, a, cc); cc=S->separator; }
  if ( flag_a ){ FILE2_print_int (fp, ay, cc); cc=S->separator; }
}

/* update sc for a newly found position */
int POS_update_sc (FILE2 *fp, POS *P, SEQ_LONG a, int k){
  SEQ *S = P->S;
  SEQ_LONG t;
  int l, flag=0;
  
  // update sc
  if ( P->flag & POS_SC_COUNT_INTERVAL ){  // counting intervals
    l = ((P->flag&POS_SC_COUNT_INTERVAL_EXTEND)? S->plen_org: 1);
    l = (k < S->plen_org? 0: k-S->plen_org)+l;
    FLOOP (t, a, a+l) if ( P->sc[t] < POS_SC_END-1 ) P->sc[t]++;
  } else if ( P->flag & POS_SC_COUNT ){  // counting similar substring
    if ( P->sc[a] < P->sc_max-1 ) P->sc[a]++; else return (0);
  } else if ( P->flag & POS_SC_MIN ){  // computing mismatch tolerance
    FLOOP (t, a, a+((P->flag&POS_SC_COUNT_INTERVAL_EXTEND)? S->plen_org: 1)){
      if ( k <= P->sc[t] + P->k_level ){ flag = 1; ENMIN (P->sc[t], k); }
    }
    return (flag);
  }
  return (1);
}

/* update str_sc for a newly found position */
int POS_update_str_sc (FILE2 *fp, POS *P, SEQ_LONG a, int k){
  SEQ *S = P->S;
  SEQ_ID sa = SEQ_get_segID (S, a);
  if ( P->flag & POS_STR_SC_COUNT ){  // counting similar stings
    if ( P->str_sc[sa] < P->str_sc_max-1 ) P->str_sc[sa]++; else return (0);
  } else if ( P->flag & POS_STR_SC_MIN ){  // computing mismatch tolerance
    if ( k <= P->str_sc[sa]+P->k_level ) ENMIN (P->str_sc[sa], k); else return (0);
  }
  return (1);
}

/**/
void POS_fprint_pair (FILE2 *fp, POS *P, SEQ_LONG a, SEQ_LONG b, int k, int core_id){
  SEQ *S = P->S;
  int flag = 1;
  SEQ_LONG t;
  
  if ( a>b ){ t = a; a = b; b = t; }
  if ( (S->flag&SEQ_OUTPUT_PAIR) && S->perr_min<S->perr_org && k<=S->perr_org-S->perr_inc ) return;

  // update sc
  SPIN_LOCK(P->I->multi_core, P->I->lock_sc);
  if ( P->flag & POS_SC_USE ){
    flag = 0;
    if ( a < S->output_pos_fprint_pair ) flag += POS_update_sc (fp, P, a, k);
    if ( b < S->output_pos_fprint_pair ) flag += POS_update_sc (fp, P, b, k);
  }
  if ( P->flag & POS_STR_SC ){
    flag = 0;
    if ( a < S->output_pos_fprint_pair ) flag += POS_update_str_sc (fp, P, a, k);
    if ( b < S->output_pos_fprint_pair ) flag += POS_update_str_sc (fp, P, b, k);
  }
  SPIN_UNLOCK(P->I->multi_core, P->I->lock_sc);

  if ( !(S->flag & SEQ_OUTPUT_PAIR) ||
      ((S->flag&SEQ_MULTI_OUTPUT) && (fp->fp==NULL)) ) return;  // no need to output pair
   //   (S->flag&SEQ_MULTI_OUTPUT) ????

  if ( flag == 0 ) return;
  if ( S->output_pos > 0 && a >= S->output_pos && b >= S->output_pos ) return; // unnecessary pair

// sort the pair when self comparison mode
  P->I->multi_outputs[core_id]++;
  if ( P->I->max_solutions>0 && P->I->multi_outputs[core_id] > P->I->max_solutions ){
    ITEMSET_last_output (P->I);
    ERROR_MES = "reached to maximum number of solutions";
    EXIT;
  }

    // print a position pair
  SEQ_fprint_position (fp, S, a, 0);
  SEQ_fprint_position (fp, S, b, S->separator);
  FILE2_print_int (fp, k, S->separator);
  
  if ( S->flag & SEQ_OUTPUT_STRING ){
    t = S->plen_org;
    if ( k>S->plen_org ) S->plen_org = k;
    SEQ_fprint_interval (fp, S, a, a, 1);
    SEQ_fprint_interval (fp, S, b, b, 1);
    S->plen_org = t;
  }
  FILE2_putc (fp, '\n');
  if ( !(S->flag&SEQ_MULTI_OUTPUT) || (fp->buf-fp->buf_org) > FILE2_BUFSIZ/2 ){
    SPIN_LOCK(P->I->multi_core, P->I->lock_output);
    FILE2_flush (fp);
    SPIN_UNLOCK(P->I->multi_core, P->I->lock_output);
  }
}

/* read the values from sc file */
void POS_read_sc (POS *P, char *fname){
  SEQ_LONG x;
  POS_SC y;
  FILE2 fp;
  FILE2_open (fp, fname, "r", EXIT);
  while (1){
    x = (SEQ_LONG)FILE2_read_int (&fp);
    y = (POS_SC)FILE2_read_int (&fp);
    if ( FILE_err&2 ) break;
    FILE2_read_until_newline ( &fp);
    P->sc[x/POS_DEFAULT_SC_SKIP] = y;
  }
}

/* read the values from sc file, for string sc */
void POS_read_str_sc (POS *P, char *fname){
  SEQ_LONG x;
  POS_SC y;
  FILE2 fp;
  FILE2_open (fp, fname, "r", EXIT);
  while (1){
    x = (SEQ_LONG)FILE2_read_int (&fp);
    y = (POS_SC)FILE2_read_int (&fp);
    if ( FILE_err&2 ) break;
    FILE2_read_until_newline ( &fp);
    P->str_sc[x] = y;
  }
}

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

/* following routines for making position lists are only for original string positions */
/* compute optimal interleave
 set interleave_len, pos_skip_a,b */
void POS_compute_interleave (SEQ *S){
  SEQ_ID i, ii, tail = S->separated_by, head = S->separated_by;
  SEQ_LONG ll=S->str_len, len1=0, len2=0, l;
  int x, v;

  if ( S->pos_skip_a > 0 ){
    S->pos_skip_b = S->interleave_len_org/S->pos_skip_a;
    goto END;
  }
  if ( S->separated_by == 0 ){ head = 1; tail = S->str_num; }
  FLOOP (i, 1, tail) len1 += S->tail[i] - S->head[i];
  FLOOP (i, head, S->str_num) len2 += S->tail[i] - S->head[i];
  if ( S->ylen > 1 ){   // matrix mode
    S->pos_skip_a = S->xlen; S->pos_skip_b = S->ylen;
//    if ( len1/S->xlen + len2/S->ylen < len1/S->ylen + len2/S->xlen )
//        SWAP_INT ( S->pos_skip_a, S->pos_skip_b);
    ll = len1/S->pos_skip_a + len2/S->pos_skip_b;
  } else {
    FLOOP (x, 1, S->interleave_len_org+1){
      l=0; v = (S->interleave_len_org)/x;
      FLOOP (ii, 1, tail){
        if ( ii < (SEQ_ID)S->matrix_by )
          l += ((S->clms[ii]-S->plen_org+x)/x) * ((S->tail[ii]-S->head[ii]+S->clms[ii]-1)/S->clms[ii]);
        else l += (S->tail[ii]-S->head[ii]-S->plen_org+x)/x;
      }
      FLOOP (ii, head, S->str_num){
        if ( ii < (SEQ_ID)S->matrix_by )
          l += (((int)((S->clms[ii]-S->plen_org+x)/x/v))*x) * ((S->tail[ii]-S->head[ii]+S->clms[ii]-1)/S->clms[ii]);
        else l += ((int)((S->tail[ii]-S->head[ii]-S->plen_org+x)/x/v))*x;
      }
      if ( x==1 || ll > l ){
        ll = l;
        S->pos_skip_a = x; S->pos_skip_b = v;
      }
    }
  }
  END:;
  S->interleave_len = S->pos_skip_a * S->pos_skip_b;
}

/* set position list to optimal positions for string comparison */
void POS_mk_pos_list (POS *P){
  SEQ *S = P->S;
  SEQ_ID i, tail = S->separated_by, head = S->separated_by;
  SEQ_LONG j, jj, jjj, skip, skip2;
  int f;
  FILE2 fp;

  POS_compute_interleave (S);
  P->pos_list = NULL;
  if ( S->separated_by==0 ){ head = 1; tail = S->str_num; }

  if ( P->fname ){
    FILE2_open (fp, P->fname,"r",EXIT);
    ARY_SCAN (P->pos_list_num, int, fp, 1); // count #positions
    if ( (P->flag) & POS_LOAD_STRPOS_PAIR ) P->pos_list_num /= 2;
    P->pos_list_end = P->pos_list_num +2;
    malloc2 (P->pos_list, P->pos_list_end, EXIT);
    FILE2_reset (&fp);
    jjj = 1; skip = 0; // string ID
    FLOOP (j, 0, P->pos_list_num){
      do {
        if ( (P->flag) & POS_LOAD_STRPOS_PAIR ){
          jjj = FILE2_read_int (&fp);
          P->pos_list[j] = FILE2_read_int (&fp) + S->head[jjj];
        } else {
          P->pos_list[j] = FILE2_read_int (&fp);
          while ( P->pos_list[j] >= skip + (S->tail[jjj] - S->head[jjj]) ){
            skip += (S->tail[jjj] - S->head[jjj]);
            jjj++;
          }
          P->pos_list[j] += S->head[jjj] - skip;
        }
      } while ((FILE_err&6) == 4);
      if (FILE_err&2) break;
    }
    FILE2_close (&fp);
    return;
  }

  while (1){
    P->pos_list_num = 0;
    FLOOP (i, 1, tail){
      if ( S->tail[i]-S->head[i] < (SEQ_LONG)MAX(S->interleave_len_org, S->plen_org) ) continue;
      skip = ((S->ylen>0)&&(i<(SEQ_ID)S->matrix_by))? S->clms[i]: S->tail[i]-S->head[i];
      ENMAX (skip, (SEQ_LONG)S->plen_org);
      for (jj=S->head[i] ; jj<S->tail[i] ; jj+=skip){  // for matrix mode, unnnecessary loop usual
        for (j=jj ; j<=MIN(jj+skip, S->tail[i])-S->plen_org ; j+=S->pos_skip_a){
          if ( P->pos_list ) P->pos_list[P->pos_list_num] = j;
          P->pos_list_num++;
        }
      }
    }
    FLOOP (i, head, S->str_num){
      f = (S->ylen==1 && i<(SEQ_ID)S->matrix_by);
      if ( S->tail[i]-S->head[i] < (SEQ_LONG)MAX(S->interleave_len_org, S->plen_org) ) continue;
      skip = (S->ylen>1&&i<(SEQ_ID)S->matrix_by)? S->clms[i]: S->pos_skip_a;
      skip2 = f? S->clms[i]: S->tail[i]-S->head[i];
      for (jjj=S->head[i] ; jjj<S->tail[i] ; jjj += skip2){
        for (j=jjj ; j<=jjj+skip2-S->plen_org ; j += skip * S->pos_skip_b){
          FLOOP (jj, j, MIN(j+skip, S->tail[i]-S->plen_org+1)){
            if ( S->separated_by==0 && jj==j ) continue; // for self comparison; avoid duplication
//printf ("ccc %d\n", jj);
            if ( f && ((jj-S->head[i])%S->clms[i])+S->plen_org > S->clms[i] )
                jj = ((int)((jj-S->head[i])/S->clms[i]+1)) * S->clms[i];  // goto next row
            if ( P->pos_list ) P->pos_list[P->pos_list_num] = jj;
            P->pos_list_num++;
          }
        }
      }
    }
    if ( P->pos_list ) break;
    P->pos_list_end = P->pos_list_num +2;
    malloc2 (P->pos_list, P->pos_list_end+2, EXIT);
  }
}

/* remove all positions from posision list which include skip_char's no less than skip_chr_ub */
void POS_rm_pos_with_ignore_chr (POS *P){
  SEQ_LONG i, ii=0;
  int j, k;
  SEQ *S = P->S;
  FLOOP (i, 0, P->pos_list_num){
    k=0; FLOOP (j, 0, S->plen_org)
        if ( S->unperm[S->s[P->pos_list[i]+j]] == P->skip_chr ) k++;
    if ( k <= P->skip_chr_ub ){
      P->pos_list[ii] = P->pos_list[i];
      ii++;
    }
  }
  P->pos_list_num = ii;
}

/* remove all repeat strings */
void POS_rm_pos_repeated_string (POS *P, int th){
  SEQ_LONG i, ii=0, p;
  int k, kmin, jj;
  SEQ_CHAR *s1, *s2;
  SEQ *S = P->S;

  FLOOP (i, 0, P->pos_list_num){
    p = P->pos_list[i]; kmin = S->plen_org;
    FLOOP (jj, 1, MIN(P->skip_repeat_check,S->plen_org/2)+1){
      for (k=0,s1=S->s+p,s2=s1+jj ; s2<S->s+p+S->plen_org ; s1++,s2++)
          if ( SEQ_wildcard_mismatch (S, *s1, *s2) ) k++;
      ENMIN ( kmin, k);
    }
    if ( kmin > th ){
      P->pos_list[ii] = P->pos_list[i];
      ii++;
    }
  }
  print_mes (S, "#repeat-string-positions= " SEQ_LONGF "\n", P->pos_list_num-ii);
  P->pos_list_num = ii;
}

/* remove all positions with small sc, or small str_sc */
/* flag&1: str_sc, o.w. sc,   flag&2: remoce one's larger than th */
void POS_reduce_pos (POS *P){
  SEQ_LONG i, ii=0, p;
  SEQ *S = P->S;
  
  FLOOP (i, 0, P->pos_list_num){
    p = P->pos_list[i];
    if ( p < S->reduced_pos ){
      if ( P->flag & POS_STR_SC ){
        if ( P->str_sc[SEQ_get_segID (S, p)] != P->init_str_sc ) continue;
      } else if ( (P->flag&POS_SC_USE) && P->sc[p] != P->init_sc ) continue;
    }
    P->pos_list[ii] = p;
    ii++;
  }
  P->pos_list_num = ii;
}

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/


/*******************************************************************************/
/* the following bucket sort routines accept shifted positions */
/* compute the 'letter' obtained by combining l characters, beginning from kth
 letter of position "pos" to (l-1)th letter, and it for matrix mode */
int SEQ_bucket_letter (SEQ *S, SEQ_LONG pos, int k, int l){
#ifdef SEQ_PPERM_MASK
  int c=0;
  SEQ_LONG x = pos&(SEQ_PPERM_MASK-1);
  int perm = pos/SEQ_PPERM_MASK;
//  for (; k<l ; k++ ) c = c*S->chr_num + S->s[x+S->pperm[perm][k]];
  for (; k<l ; k++ ) c = c*S->chr_num + S->s[x+S->pperm[perm][k]];
#else
  SEQ_CHAR *x, *x_end = &S->s[pos+l];
  int c=S->s[pos+k];
/*  for (x=&S->s[pos+k+1] ; x<x_end-2 ; x+=3 )
     c = c* S->chr_num3 + *x * S->chr_num2 + *(x+1) * S->chr_num + *(x+2);
  if ( x<x_end ){
    c = c*S->chr_num + *x;
    if ( x<x_end-1 )  c = c*S->chr_num + *(x+1);
  }
*/
  FLOOP (x, &S->s[pos+k+1], x_end) c = c*S->chr_num + *x;
  
#endif
  return (c);
}
int SEQ_bucket_letter_matrix (SEQ *S, SEQ_LONG pos, int k, int l){
  int c=0, i, clms = S->clms[(S->flag&SEQ_MULTI_STRING)? SEQ_get_substID (S, pos) :1];
  SEQ_CHAR *x = &S->s[ pos + ((int)(k/S->xlen))*clms + (k%S->xlen)];
  FLOOP (i, k, l){
    c = c*S->chr_num + *x;
    i++;
    if ( i%S->xlen == 0 ) x += clms - S->xlen;
  }
  return (c);
}

void SEQ_bucket_sort (SEQ *S, SEQ_LONG *pos, SEQ_LONG t, int k, int l, QUEUE *jump, SEQ_LONG *cnt, SEQ_LONG *cnt2){
  int c, cc;
//  int bc;
//  SEQ_LONG buf[1024];
  SEQ_LONG  j, p=0, pp=0, *y, *yy;
//  SEQ_LONG *cnt=S->let_cnt[dep], *cnt2=S->let_cnt[dep+1], *po, *po_end=pos+t;
  SEQ_LONG *po, *po_end=pos+t;
  QUEUE_INT *z, *z_end;

//SEQ_LONG i;
//for ( i=0 ; i<t ; i++ ) if ( pos[i] > S->str_len ) printf ("okasii %d %d\n", pos[i], t);

  jump->t = 0;

    // counting #positions having the same hash-key

  z=jump->v; FLOOP (po, pos, po_end){
    c = S->ylen>1? SEQ_bucket_letter_matrix (S,*po,k,l): SEQ_bucket_letter (S,*po,k,l);
    if ( cnt[c] == 0 ){ *z=c; z++; }
    cnt[c]++;
//printf ("cnt[5107] %d\n", cnt[5107]);
  }
    // loop is opened, to 3 same-operations

    // compute the start position of each group by taking sum of the sizes of previous groups
  z_end = z;
  jump->t = (QUEUE_ID)(z_end - jump->v);
  j=0; for (z=jump->v ; z<z_end ; z++){
    cnt2[*z] = j; j += cnt[*z]; cnt[*z] = j;
  }

    // classify the positions into groups
  FLOOP (z, jump->v, z_end){
    FLOOP (y, pos+cnt2[*z], pos+cnt[*z]){
      p = *y;
      while (1){
          // move position to the current tail of its class, after 
        c = S->ylen>1? SEQ_bucket_letter_matrix(S, p, k, l): SEQ_bucket_letter(S, p, k, l);
        if ( c == *z ){ *y = p; break; }
        yy = pos + cnt2[c];
        pp = *yy;

        cc = S->ylen>1? SEQ_bucket_letter_matrix(S, pp, k, l): SEQ_bucket_letter(S, pp, k, l);
        cnt2[c]++;
        *yy = p;

        if ( cc == *z ){ *y = pp; break; }
        yy = pos + cnt2[cc];
        p = *yy;
        cnt2[cc]++;
        *yy = pp;
      }
    }
    cnt2[*z] = 0;
  }
}

void SEQ_bucket_sort_ (SEQ *S, SEQ_LONG *pos, SEQ_LONG t, int k, int l, QUEUE *jump, SEQ_LONG *cnt, SEQ_LONG *cnt2){
  int c, cc, ccc;
//  int bc;
//  SEQ_LONG buf[1024];
  SEQ_LONG  j, jj, jjj, p=0, pp=0, *y, *yy;
//  SEQ_LONG *cnt=S->let_cnt[dep], *cnt2=S->let_cnt[dep+1], *po, *po_end=pos+t;
  SEQ_LONG *po, *po_end=pos+t;
  QUEUE_INT *z, *z_end;

//SEQ_LONG i;
//for ( i=0 ; i<t ; i++ ) if ( pos[i] > S->str_len ) printf ("okasii %d %d\n", pos[i], t);

  jump->t = 0;

    // counting #positions having the same hash-key
  z=jump->v; for (po=pos ; po<po_end-2 ; po+=3){
    c = S->ylen>1? SEQ_bucket_letter_matrix (S,*po,k,l): SEQ_bucket_letter (S,*po,k,l);
    cc = S->ylen>1? SEQ_bucket_letter_matrix (S,*(po+1),k,l): SEQ_bucket_letter (S,*(po+1),k,l);
    ccc = S->ylen>1? SEQ_bucket_letter_matrix (S,*(po+2),k,l): SEQ_bucket_letter (S,*(po+2),k,l);
    if ( cnt[c] == 0 ){ *z=c; z++; }
    cnt[c]++;
    if ( cnt[cc] == 0 ){ *z=cc; z++; }
    cnt[cc]++;
    if ( cnt[ccc] == 0 ){ *z=ccc; z++; }
    cnt[ccc]++;
  }
  if ( po<po_end ){
    c = S->ylen>1? SEQ_bucket_letter_matrix (S,*po,k,l): SEQ_bucket_letter(S, *po, k, l);
    if ( cnt[c] == 0 ){ *z=c; z++; }
    cnt[c]++;
    po++;
    if ( po<po_end ){
      c = S->ylen>1? SEQ_bucket_letter_matrix (S,*po,k,l): SEQ_bucket_letter(S, *po, k, l);
      if ( cnt[c] == 0 ){ *z=c; z++; }
      cnt[c]++;
    }
  }
    // loop is opened, to 3 same-operations

    // compute the start position of each group by taking sum of the sizes of previous groups
  z_end = z;
  jump->t = (QUEUE_ID)(z_end - jump->v);
  j=0; for (z=jump->v ; z<z_end-2 ; z+=3){
    cnt2[*z] = j,
    jj = cnt[*z]+j, jjj = jj+cnt[*(z+1)], j = jjj+cnt[*(z+2)]; 
    cnt2[*(z+1)] = cnt[*z] = jj;
    cnt2[*(z+2)] = cnt[*(z+1)] = jjj;
    cnt[*(z+2)] = j;
  }
  if ( z < z_end ){
    cnt2[*z] = j, cnt[*z] += j;
    if ( z < z_end-1 ){
      cnt2[*(z+1)] = cnt[*z], cnt[*(z+1)] += cnt[*z];
    }
  }

    // classify the positions into groups
  FLOOP (z, jump->v, z_end){
    FLOOP (y, pos+cnt2[*z], pos+cnt[*z]){
      p = *y;
      while (1){
          // move position to the current tail of its class, after 
        c = S->ylen>1? SEQ_bucket_letter_matrix(S, p, k, l): SEQ_bucket_letter(S, p, k, l);
        if ( c == *z ){ *y = p; break; }
        yy = pos + cnt2[c];
        pp = *yy;

        cc = S->ylen>1? SEQ_bucket_letter_matrix(S, pp, k, l): SEQ_bucket_letter(S, pp, k, l);
        cnt2[c]++;
        *yy = p;

        if ( cc == *z ){ *y = pp; break; }
        yy = pos + cnt2[cc];
        p = *yy;
        cnt2[cc]++;
        *yy = pp;
      }
    }
    cnt2[*z] = 0;
  }
}

void SEQ_bucket_sort_by_substID (POS *P, SEQ_LONG *pos, SEQ_LONG t){
  SEQ *S = P->S;
  SEQ_ID c, cc;
  SEQ_LONG  i, j=0, p, pp, x, xx;
  QUEUE_INT *z = P->IDjump.v, *z_end;

  P->IDjump.t = 0;
//printf ("ccc %d %d %d  (%d->%d)\n", pos-S->pos_list, t, dep, k, l);
  FLOOP (i, 0, t){
    c = SEQ_get_substID (S, SEQ_get_pos(pos[i]));
    if ( P->IDcnt[c] == 0 ){ *z=c; z++; }
    P->IDcnt[c]++;
  }
  z_end = z;
  P->IDjump.t = (QUEUE_ID)(z_end - P->IDjump.v);
  FLOOP (z, P->IDjump.v, z_end){
//QUEUE_FE_LOOP_ ( P->jump[dep], i, c ){
    P->IDcnt2[*z] = j;
    j += P->IDcnt[*z];
    P->IDcnt[*z] = j;
  }
  FLOOP (z, P->IDjump.v, z_end){
    FLOOP (x, P->IDcnt2[*z], P->IDcnt[*z]){
      p = pos[x];
      while (1){
        cc = SEQ_get_substID (S, SEQ_get_pos(p));
        if ( cc == *z ) break;
        xx = P->IDcnt2[cc];
//        printf ("pos[xx]=%d, cnt %d, cnt+1 %d\n", pos[xx], S->let_cnt[dep][cc], S->let_cnt[dep+1][cc]);
        pp = pos[xx];
        P->IDcnt2[cc]++;
        pos[xx] = p;
        p = pp;
      }
      pos[x] = p;
    }
    P->IDcnt2[*z] = 0;
  }
}

/* partition the given position list into two groups of ID>=separated_by, and ID<... */
SEQ_LONG SEQ_partition_by_substID (SEQ *S, SEQ_LONG *pos, SEQ_LONG t){
  SEQ_LONG i=0, x;
  while (1){
    while (1){
      if ( pos[i] >= S->separated_pos ) break;
      i++;
      if ( i==t ) return(i);
    }
    while (1){
      t--;
      if ( i==t ) return (i);
      if ( pos[t] < S->separated_pos ) break;
    }
    x = pos[i]; pos[i] = pos[t]; pos[t] = x;
  }
  i++;
  if ( i==t ) return (i);
}
SEQ_LONG SEQ_count_by_substID (SEQ *S, SEQ_LONG *pos, SEQ_LONG t){
  SEQ_LONG i, j=0;
  FLOOP (i, 0, t) if ( pos[i] < S->separated_pos ) j++;
  return (j);
}

/* paritition the positions by output_by, separated_by, and sc */
SEQ_LONG POS_partition_by (POS *P, SEQ_LONG *pos, SEQ_LONG t){
  SEQ *S = P->S;
  SEQ_LONG i=0, x;

  while (1){
    while (1){
      if ( pos[i] >= P->sep_pos ) break;
      if ( (P->flag & POS_STR_SC_COUNT) && P->str_sc[ SEQ_get_segID (S, pos[i])] >= P->partition_th) break;
      if ( (P->flag & POS_STR_SC_MIN) && P->str_sc[ SEQ_get_segID (S, pos[i])] < P->partition_th ) break;
      if ( (P->flag & POS_SC_COUNT) && P->sc[pos[i]/POS_DEFAULT_SC_SKIP] >= P->partition_th) break;
      if ( (P->flag & POS_SC_MIN) && P->sc[pos[i]/POS_DEFAULT_SC_SKIP] < P->partition_th) break;
      i++;
      if ( i==t ) return (i);
    }
    while (1){
      t--;
      if ( i==t ) return (i);
      if ( pos[t] < P->sep_pos ){
        if ( P->flag & POS_STR_SC_COUNT){
          if ( P->str_sc[ SEQ_get_segID (S, pos[i])] >= P->partition_th) continue;
        } else if ( (P->flag & POS_STR_SC_MIN) && P->str_sc[ SEQ_get_segID (S, pos[i])] < P->partition_th ) continue;
        if ( P->flag & POS_SC_COUNT ){
          if ( P->sc[pos[i]/POS_DEFAULT_SC_SKIP] < P->partition_th ) continue;
        } else if ( (P->flag & POS_SC_MIN) && P->sc[pos[i]/POS_DEFAULT_SC_SKIP] >= P->partition_th) continue;
        break;
      }
    }
    x = pos[i]; pos[i] = pos[t]; pos[t] = x;
  }
}
SEQ_LONG POS_count_by (POS *P, SEQ_LONG *pos, SEQ_LONG t){
  SEQ *S = P->S;
  SEQ_LONG i, j=0;
  FLOOP (i, 0, t){
    if ( pos[i] >= P->sep_pos ) continue;
    if ( (P->flag & POS_STR_SC_COUNT) && P->str_sc[ SEQ_get_segID (S, pos[i])] >= P->partition_th) continue;
    if ( (P->flag & POS_STR_SC_MIN) && P->str_sc[ SEQ_get_segID (S, pos[i])] < P->partition_th ) continue;
    if ( (P->flag & POS_SC_COUNT) && P->sc[pos[i]/POS_DEFAULT_SC_SKIP] >= P->partition_th) continue;
    if ( (P->flag & POS_SC_MIN) && P->sc[pos[i]/POS_DEFAULT_SC_SKIP] < P->partition_th) continue;
    j++;
  }
  return (j);
}

/**********************************************************************************/
/**********************************************************************************/
/**********************************************************************************/

/* compute #mismatches in the remaining blocks
  if the mismatches > SS->perr, return SS->plen
  if the current combination of the blocks is not canonical for the given pair, return S->plen */
#ifdef SEQ_PPERM_MASK
int POS_remain_mismatch (POS *P, SEQ_LONG x, SEQ_LONG y, int block, int block_end){
  SEQ *S = &P->S;
  int j, k=0, kk, l, ll;
  SEQ_CHAR *s1, *s2, *px=S->s+x, *py=S->s+y;
  QUEUE_INT *z;

  int px = x/SEQ_PPERM_MASK, py = y/SEQ_PPERM_MASK;
  int *ppx = S->pperm[px], *ppy=S->pperm[py];
  x = x&(SEQ_PPERM_MASK-1);
  y = y&(SEQ_PPERM_MASK-1);
  s1 = S->s+x; s2 = S->s+y;

    // compute the Ham. dist of previous blocks and return if there is an exact match
  MQUE_FLOOP (P->remain_block, z){
    kk = 0;
    FLOOP (j, P->ppartition.v[*z *2], P->ppartition.v[*z *2+1]){
      if ( *(s1+*(ppx+j)) != *(s2+*(ppy+j)) ) k++,kk++;
      if ( k > S->perr ) return (S->plen_org);
    }
    if ( kk==0 ) return (S->plen_org);
  }
  if ( S->wildcard > 0 ){
    k=0; MQUE_FLOOP (P->remain_block, z){
      FLOOP (j, P->ppartition.v[*z*2], P->ppartition.v[*z*2+1])
          k += SEQ_wildcard_mismatch (S, *(s1+*(ppx+j)), *(s2+*(ppy+j)));
  }
    // compute the Ham. dist of latter blocks
// modification, same as the following, is needed
  for ( ; block<block_end ; block++ ){
    FLOOP (j, P->ppartition.v[block*2], P->ppartition.v[block*2+1]){
      k += SEQ_wildcard_mismatch (S, *(s1+*(ppx+j)), *(s2+*(ppy+j)) );
      if ( k > S->perr ) return (S->plen_org);
    }
  }
//SEQ_fprint_substring (stdout, S, x, S->plen); printf ("   ");
//SEQ_fprint_substring (stdout, S, y, S->plen); printf ("   %d\n",k);
  
  // compute the Ham. dist of preceding chunks and following chunks
  // if there is a preceding block with HamDist<=S->perr, return false from the canonicity

  px -= P->chunk_head; py -= P->chunk_head;
  FLOOP (j, 0, P->chunk){
    if ( j == P->cur_chunk ){
      if ( PP->problem & SACHICA_ONE_CHUNK_MATCH ) break;  // for "output if one chunk matches"
      else continue;
    }
    l = P->chunk_pos[j]; ll = P->chunk_pos[j+1];
    for (s1=px+l,s2=py+l ,kk=0 ; s1<px+ll ; s1++, s2++)
      kk += SEQ_wildcard_mismatch (S, *s1, *s2);
    if ( kk<=S->perr && j<P->cur_chunk ) return (S->plen_org);
    k += kk;
    if ( k > S->perr_org && !(P->flag & POS_ONE_CHUNK_MATCH) ) return (S->plen_org); // PP->problem... is for "output if one chunk matches"
  }
  return (k);
}

#else
/* compute #mismatches in the remaining blocks
  if the mismatches > S->perr, return S->plen
  if the current combination of the blocks is not canonical for the given pair, return S->plen */
int POS_remain_mismatch (POS *P, SEQ_LONG x, SEQ_LONG y, int block, int block_end){
  SEQ *S = P->S;
  int j, k=0, kk, l, ll;
  SEQ_CHAR *s1, *s2, *px=S->s+x, *py=S->s+y, *ss;
  QUEUE_INT *z;

    // compute the Ham. dist of previous blocks and return if there is an exact match

  MQUE_FLOOP (P->remain_block, z){
    j = P->ppartition.v[*z*2]; l = P->ppartition.v[*z*2+1];
/*
#ifdef USE_SIMD
    __m128i u1, u2;
    char ss[16];
    l -= j;
    for (s1=px+j,s2=py+j ; l>16 ; s1+=16,s2+=16,l-=16){
      u1 = _mm_loadu_si128 (s1);
      u2 = _mm_loadu_si128 (s2);
      u1 = mm_cmpeq_epi8 (u1, u2);
      u2 = _mm_setzero_si128 ();
      _mm_sad_epu8(u1, u2);
      _mm_storeu_si128(ss, u1);
      k += ss[0] + ss[8];
    }
  u1 = _mm_loadu_si128 (s1);
  u2 = _mm_loadu_si128 (s2);
  _mm_srli_si128(u1, 16-l);
  _mm_srli_si128(u2, 16-l);
  u1 = mm_cmpeq_epi8 (u1, u2);
  u2 = _mm_setzero_si128 ();
  _mm_sad_epu8(u1, u2);
  _mm_storeu_si128(ss, u1);
  k += ss[0] + ss[8];
#else
*/
    kk = 0;
    ss = px + l-1;
    for (s1=px+j,s2=py+j ; s1<ss ; s1++,s2++){
      if ( *s1 != *s2 ){ kk++; if ( ++k > S->perr ) return (S->plen_org); }
      s1++; s2++;
      if ( *s1 != *s2 ){ kk++; if ( ++k > S->perr ) return (S->plen_org); }
    }
    if ( s1 < ss+1 ) if ( *s1 != *s2 ){ kk++; if ( ++k > S->perr ) return (S->plen_org); }
    if ( kk==0 ) return (S->plen_org);
  }
//#endif
  if ( S->wildcard > 0 ){
    k=0; MQUE_FLOOP (P->remain_block, z){
      j = P->ppartition.v[*z*2];
      for (s1=px+j,s2=py+j ; s1<px+P->ppartition.v[*z*2+1] ; s1++,s2++)
          k += SEQ_wildcard_mismatch (S, *s1, *s2);
    }
  }
    // compute the Ham. dist of latter blocks

  if ( block<block_end ){
    j = P->ppartition.v[block*2];
    l = P->ppartition.v[block_end*2-1];
    for (s1=px+j,s2=py+j ; s1<px+l ; s1++,s2++){
      k += SEQ_wildcard_mismatch (S, *s1, *s2);
      if ( k > S->perr ) return (S->plen_org);
    }
  }

//SEQ_fprint_substring (stdout, SS, x, SS->plen); printf ("   ");
//SEQ_fprint_substring (stdout, SS, y, SS->plen); printf ("   %d\n",k);
  
  // compute the Ham. dist of preceding chunks and following chunks
  // if there is a preceding block with HamDist<=S->perr, return false from the canonicity

  px -= P->chunk_head; py -= P->chunk_head;
  FLOOP (j, 0, P->chunk){
    if ( j == P->cur_chunk ){
      if ( P->flag & POS_ONE_CHUNK_MATCH ) break;  // for "output if one chunk matches"
      else continue;
    }
    l = P->chunk_pos[j]; ll = P->chunk_pos[j+1];
    for (s1=px+l,s2=py+l ,kk=0 ; s1<px+ll ; s1++, s2++)
      kk += SEQ_wildcard_mismatch (S, *s1, *s2);
    if ( kk<=S->perr && j<P->cur_chunk ) return (S->plen_org);
    k += kk;
    if ( k > S->perr_org && !(P->flag & POS_ONE_CHUNK_MATCH) ) return (S->plen_org); // SA->problem... is for "output if one chunk matches"
  }
  return (k);
}

#endif

/* compute the mismatches in other chunks */
int POS_check_chunk_ (POS *P, SEQ_CHAR *px, SEQ_CHAR *py, int k){
  int kk, j=0, kkk;
  SEQ *S = P->S;
  SEQ_CHAR *s1, *s2, *s_end;

  px -= P->chunk_head; py -= P->chunk_head;
  s1 = px; s2 = py;

  FLOOP (j, 0, P->cur_chunk){
    s_end = px + P->chunk_pos[j+1];
//    for (kk=S->cost[*s1][*s2],s1++,s2++ ; s1<s_end ; s1++,s2++)
//        kk += S->cost[*s1][*s2];
    for (kk=S->cost[*s1][*s2],s1++,s2++ ; s1<s_end-3 ; s1++,s2++){
      kkk = S->cost[*s1][*s2];
      s1++; s2++; kk += S->cost[*s1][*s2];
      s1++; s2++; kkk += S->cost[*s1][*s2];
      s1++; s2++; kk += S->cost[*s1][*s2];
      kk += kkk;
    }
    if ( s1<s_end-1 ){
      if ( s1<s_end-2 ){ kk+=S->cost[*s1][*s2]; s1++; s2++; }
      kk+=S->cost[*s1][*s2]; s1++; s2++;
      kk+=S->cost[*s1][*s2]; s1++; s2++;
    } else if ( s1<s_end ){ kk+=S->cost[*s1][*s2]; s1++; s2++; }
    if ( kk<=S->perr ) return (S->plen_org);
    k += kk;
    if ( k > S->perr_org && !(P->flag & POS_ONE_CHUNK_MATCH) ) return (S->plen_org); // PP->problem... is for "output if one chunk matches"
  }
  if ( P->flag & POS_ONE_CHUNK_MATCH ) return (k);  // for "output if one chunk matches"
  if ( P->cur_chunk == P->chunk-1 ) goto END;

  s1 = px+P->chunk_pos[P->cur_chunk+1]; s2 = py+P->chunk_pos[P->cur_chunk+1];
  s_end = px + P->chunk_pos[P->chunk+1];

//  for (; s1<s_end ; s1++, s2++) k += S->cost[*s1][*s2];

  for ( ; s1<s_end-3 ; s1++, s2++){
    kkk = S->cost[*s1][*s2];
    s1++; s2++; k += S->cost[*s1][*s2];
    s1++; s2++; kkk += S->cost[*s1][*s2];
    s1++; s2++; k += S->cost[*s1][*s2];
    k += kkk;
  }
  if ( s1<s_end-1 ){
    if ( s1<s_end-2 ){ k+=S->cost[*s1][*s2]; s1++; s2++; }
    k+=S->cost[*s1][*s2]; s1++; s2++;
    k+=S->cost[*s1][*s2]; s1++; s2++;
  } else if ( s1<s_end ){ k+=S->cost[*s1][*s2]; s1++; s2++; }
  
  END:;
  if ( k > S->perr_org ) return (S->plen_org); // too many mismatches
  return ( k);
}

/* compute #mismatches in the remaining blocks
  if the mismatches > S->perr, return S->plen
  if the current combination of the blocks is not canonical for the given pair, return S->plen */
/* for using S->cost */
int POS_remain_mismatch_ (POS *P, SEQ_LONG x, SEQ_LONG y, int block, int block_end){
  SEQ *S = P->S;
  int j, k=0, kk;
  SEQ_CHAR *s1, *s2, *px=S->s+x, *py=S->s+y;
  QUEUE_INT *z;

    // compute the Ham. dist of previous blocks and return if there is an exact match
  MQUE_FLOOP (P->remain_block, z){
    kk = 0;
    j = P->ppartition.v[*z*2];
//    for (j=S->ppartition.v[e],s2=&S->s[y+j] ; j<S->ppartition.v[e+1] ; j++){
    for (s1=px+j,s2=py+j ; s1<px+P->ppartition.v[*z*2+1] ; s1++,s2++){
      if ( *s1 != *s2 ) kk++;
      k += S->cost[*s1][*s2];
      if ( k > S->perr ) return (S->plen_org);
    }
    if ( kk==0 ) return (S->plen_org);
  }
    // compute the Ham. dist of latter blocks
  for ( ; block<block_end ; block++ ){
    j = P->ppartition.v[block*2];
    for (s1=px+j,s2=py+j ; s1<px+P->ppartition.v[block*2+1] ; s1++,s2++){
      k += S->cost[*s1][*s2];
      if ( k > S->perr ) return (S->plen_org);
    }
  }

//  for (s1=px+S->plen,s2=py+S->plen; s1>px+*(S->ppartition.v+block*2) ; ){
//    s1--; s2--; k += S->cost[*s1][*s2];
//    if ( k > S->perr ) return (SS->plen_org);
//  }
  return (POS_check_chunk_ (P, px, py, k));
}

/* extract maximum length substring continuously includes few ignoring char's */
/*
void SEQ_extract_valid_string (SEQ *S, int ){
  int k;
  SEQ_ID i;
  SEQ_LONG j, jj, m=0, ml=0, mr=0;
  FLOOP (i, 1, S->str_num){
    for (k=0,j=jj=S->head[i] ; j<S->head[i]+S->plen_org ; j++)
        if ( S->unperm[S->s[j]] == S->skip_chr ) k++;
    while (1){
      if ( k > S->skip_chr_ub ){
        if ( j-jj > m ){ ml=jj, mr=j; m=jj-j; }
        jj=j+1;
      }
      if ( j>=S->tail[i]-S->plen_org ) break;
      if ( S->unperm[S->s[j-S->plen_org]] == S->skip_chr ) k--;
      if ( S->unperm[S->s[j]] == S->skip_chr ) k++;
      j++;
    }
    for (jj=0,j=ml ; j<mr ; j++,jj++) S->s[jj] = S->s[j];
    S->tail[i] = mr - ml;
  }
}
*/



#endif


